<h2 class="display-2">Welcome to the NZ ORCID Hub</h2>
<p>Welcome!</p>
<p>{% if invitation %} {{ invitation.org.name }} {% elif org_name %} {{org_name}} {% else %} Your organisation{% endif
  %} has requested you to grant them 'PERSON/UPDATE' permission,
  which means {% if invitation %} {{ invitation.org.name }} {% elif org_name %} {{org_name}} {% else %} Your
  organisation{% endif %} can read and write other information about you (other names, researcher-urls, keywords, etc.)
  to your ORCID record, </p>
<p>If you want this to happen you'll need to give them permission...</p>
<p>Please click on the link below to login on ORCIDHUB via ORCID<br>
  <a href="{{ invitation_url }}">{{ invitation_url }}</a><br>
<p>
  Once you click this link, you'll be taken to ORCID to sign in to, or create, your ORCID record.
  Once you've granted permission, you'll be automatically taken to the Hub.
  Note no further action from you is needed.</p>
